# INTELLECTUAL PROPERTY RIGHT STATEMENT
# 
# The original source material is obtained from Christophe Seux, Class room,
# Blender project, url: https://www.blender.org/download/demo-files/, and 
# according to that website, the Classroom material is subject to the CC0 
# license.
# 
# That source material has been processed by Philips. As regards the processed
# material, and the way the processed material is obtained from the source 
# material, intellectual property rights are owned by Koninklijke Philips N.V. 
# These processed materials may only be used for the purpose of developing,
# testing and promulgating technology standards as well as for academic 
# purposes. Koninklijke Philips N.V. makes no warranties with respect to the 
# processed materials and the processing method, and expressly disclaims any 
# warranties regarding their fitness for any purpose.

import os
import bpy
import math
import json
from mathutils import *

#
# INPUT PARAMETERS FROM ENVIRONMENT VARIABLES
#

# The file with the camera parameters (e.g. ClassroomVideo.json)
camparamsfile = os.getenv('camparamsfile')

# The camera of which to take the position (e.g. v0)
viewid = os.getenv('viewid')

# Specify the number of threads for Blender
nthreads = int(os.getenv('nthreads'))

# Specify the number of samples per pixel
nsamples = int(os.getenv('nsamples'))

#
# DERIVED PARAMETERS
#

# Set camera aperture size:
if nsamples == 1:
	aperture_size = 0. 		# For depth maps (special case)
else:
	aperture_size = 0.002	# Camera with depth of field for texture raytracing

# Load the camera parameters
with open(camparamsfile) as stream:
	camparams = json.load(stream)

# Get the viewpoint
for cam in camparams['cameras']:
	if cam['Name'] == viewid:
		# Change coordinate system: OMAF/3DoF+/W6DoF --> Blender
		viewpoint = Vector([-cam['Position'][1], cam['Position'][0], cam['Position'][2]])
		print('viewpoint:', viewpoint)
		break
if not viewpoint:
	raise RuntimeError('Could not find viewpoint in camera parameters file')

#
# TRANSFORM THE BLENDER PROJECT
#

C = bpy.context

# Make sure the blinds are loaded
if not C.user_preferences.system.use_scripts_auto_execute:
	raise RuntimeError('Set: File Menu / User Preferences / File / Auto Run Python Scripts')

# Set number of allowed threads
C.scene.render.threads_mode = 'FIXED'
C.scene.render.threads = nthreads

# Set render resolution
C.scene.render.resolution_x = 4096
C.scene.render.resolution_y = 2048
C.scene.render.resolution_percentage = 100

# Non-zero camera aperture creates a natural depth of field effect
C.scene.camera.data.cycles.aperture_size = aperture_size

# Turn on Stereo 3D (top-bottom) render output
C.scene.render.use_multiview = True
C.scene.render.image_settings.views_format = 'STEREO_3D'
C.scene.render.image_settings.stereo_3d_format.display_mode = 'TOPBOTTOM'

# Set camera lens to equirectangular panorama
C.scene.camera.data.type = 'PANO'
C.scene.camera.data.cycles.panorama_type = 'EQUIRECTANGULAR'

# Set stereo camera to parallel with 0 separation and turning on spherical stereo
# NOTE: Turning of use_multiview instead results in lighting issues
C.scene.camera.data.stereo.convergence_mode = 'PARALLEL'
C.scene.camera.data.stereo.interocular_distance = 0. 
C.scene.camera.data.stereo.use_spherical_stereo = True

# Set the number of samples
# NOTE: Output is two views top-bottom with different noise pattern. The result should be averaged.
C.scene.cycles.samples = (nsamples + 1) // 2

# Set camera scale
C.scene.camera.scale = Vector([1, 1, 1])

# Turn on the Mist pass to calculate an anti-aliased depth map in [0, 10] m range
C.scene.render.layers[0].use_pass_mist = True
C.scene.world.mist_settings.start = 0.
C.scene.world.mist_settings.depth = 10.
C.scene.world.mist_settings.falloff = 'LINEAR'

# Remove all vignetting nodes
for name in ['Bokeh Image', 'Alpha Over.001', 'Invert.002', 'Scale', 'Scale.001', 'Invert.003', 'Blur.003']:
	node = C.scene.node_tree.nodes[name]
	C.scene.node_tree.nodes.remove(node)
	
def remove_reconnect(name):
	node = C.scene.node_tree.nodes[name]
	insocket = node.inputs['Image']
	outsocket = node.outputs['Image']
	sourcenode = None
	sourcesocket = None
	sinknode = None
	sinksocket = None
	
	for link in C.scene.node_tree.links:
		if link.to_node == node and link.to_socket == insocket:
			sourcenode = link.from_node
			sourcesocket = link.from_socket
			
		if link.from_node == node and link.from_socket == outsocket:
			sinknode = link.to_node
			sinksocket = link.to_socket
			
	C.scene.node_tree.nodes.remove(node)
	C.scene.node_tree.links.new(sourcesocket, sinksocket)

# Remove lens distortion and dispersion (relinking graph)
remove_reconnect('Lens Distortion')
remove_reconnect('Lens Distortion.001')

# Image output
C.scene.render.filepath = '//{}_{}spp/Color####.exr'.format(viewid, nsamples)
C.scene.render.image_settings.file_format = 'OPEN_EXR'
C.scene.render.image_settings.color_depth = '32'

# Mist output
outputfile = C.scene.node_tree.nodes.new('CompositorNodeOutputFile')
outputfile.file_slots[0].path = 'Mist'
outputfile.base_path = '//{}_{}spp'.format(viewid, nsamples)
outputfile.location = Vector((1521, 488))
outputfile.format.file_format = 'OPEN_EXR'
outputfile.format.color_depth = '32'
sourcesocket = C.scene.node_tree.nodes['Render Layers.003'].outputs['Mist']
C.scene.node_tree.links.new(sourcesocket, outputfile.inputs[0])

# Remove existing F-curve
for item in [C.scene, C.scene.camera]:
	for fcurve in item.animation_data.action.fcurves:
		item.animation_data.action.fcurves.remove(fcurve)
	
# Animated seed
# When rendering out multiple times, change the inital seed
# See: https://developer.blender.org/rB7c06190882a1
C.scene.cycles.seed = 42
C.scene.cycles.use_animated_seed = True

# Animation (10 seconds @ 30 Hz)
C.scene.frame_start = 1
C.scene.frame_end = 300

# Start keyframing
loc0 = Vector([-1.3, -0.6, 1.5])
loc1a = Vector([0.8, -0.6, 1.7]) + viewpoint
loc1b = Vector([0.8, 0, 1.7]) + viewpoint
loc2 = Vector([2.6, 2.4, 1.3]) - viewpoint
rot0 = Vector([math.pi/2, 0., -math.pi/2])
rot1 = Vector([math.pi/2, 0., 0.])
rot2 = Vector([math.pi/2, 0., math.pi])
store_armatures = list(map(lambda name: C.scene.objects[name], ['store_armature.007', 'store_armature.006', 'store_armature.005', 'store_armature.004', 'store_armature.003', 'store_armature.002', 'store_armature.001', 'store_armature']))
blind_open, blind_closed = Vector([1., 0., 0., -1./3.]), Vector([1., 0., 0., 1./3.])
blind_down, blind_up = 0., 1.6

print('loc0:', loc0)
print('loc1a:', loc1a)
print('loc1b:', loc1b)
print('loc2:', loc2)

# ... Frame 1 (0 s)
C.scene.frame_set(frame = 1)
C.scene.camera.location = loc1a
C.scene.camera.keyframe_insert(data_path = "location")
C.scene.camera.rotation_euler = rot1
C.scene.camera.keyframe_insert(data_path = "rotation_euler")

for armature in store_armatures:
	armature.pose.bones['Bone'].location[2] = blind_down
	armature.pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)
	armature.pose.bones['Bone.002'].rotation_quaternion = blind_open
	armature.pose.bones['Bone.002'].keyframe_insert(data_path = "rotation_quaternion", index = 3)

# ... Frame 31 (1 s)
C.scene.frame_set(frame = 31)
store_armatures[0].pose.bones['Bone'].location[2] = blind_down
store_armatures[0].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)
	
# ... Frame 61 (2 s)
C.scene.frame_set(frame = 61)
for armature in store_armatures:
	armature.pose.bones['Bone.002'].rotation_quaternion = blind_closed
	armature.pose.bones['Bone.002'].keyframe_insert(data_path = "rotation_quaternion", index = 3)

store_armatures[1].pose.bones['Bone'].location[2] = blind_down
store_armatures[1].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)
	
# ... Frame 91 (3 s)
C.scene.frame_set(frame = 91)
store_armatures[0].pose.bones['Bone'].location[2] = blind_up
store_armatures[0].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)
store_armatures[2].pose.bones['Bone'].location[2] = blind_down
store_armatures[2].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)

# ... Frame 106 (3 1/2 s)
C.scene.frame_set(frame = 106)
C.scene.camera.keyframe_insert(data_path = "location")

# ... Frame 121 (4 s)
C.scene.frame_set(frame = 121)
store_armatures[1].pose.bones['Bone'].location[2] = blind_up
store_armatures[1].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)
store_armatures[3].pose.bones['Bone'].location[2] = blind_down
store_armatures[3].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)

# ... Frame 150 (4 29/30 s)
C.scene.frame_set(frame = 150)
C.scene.camera.location = loc1b
C.scene.camera.keyframe_insert(data_path = "location")
C.scene.camera.keyframe_insert(data_path = "rotation_euler")

# ... Frame 151 (5 s)
C.scene.frame_set(frame = 151)
C.scene.camera.location = loc2
C.scene.camera.keyframe_insert(data_path = "location")
C.scene.camera.rotation_euler = rot2
C.scene.camera.keyframe_insert(data_path = "rotation_euler")
store_armatures[2].pose.bones['Bone'].location[2] = blind_up
store_armatures[2].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)
store_armatures[4].pose.bones['Bone'].location[2] = blind_down
store_armatures[4].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)

# ... Frame 181 (6 s)
C.scene.frame_set(frame = 181)
store_armatures[3].pose.bones['Bone'].location[2] = blind_up
store_armatures[3].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)
store_armatures[5].pose.bones['Bone'].location[2] = blind_down
store_armatures[5].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)

# ... Frame 211 (7 s)
C.scene.frame_set(frame = 211)
store_armatures[4].pose.bones['Bone'].location[2] = blind_up
store_armatures[4].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)
store_armatures[6].pose.bones['Bone'].location[2] = blind_down
store_armatures[6].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)

# ... Frame 241 (8 s)
C.scene.frame_set(frame = 241)
store_armatures[5].pose.bones['Bone'].location[2] = blind_up
store_armatures[5].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)
store_armatures[7].pose.bones['Bone'].location[2] = blind_down
store_armatures[7].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)

# ... Frame 271 (9 s)
C.scene.frame_set(frame = 271)
store_armatures[6].pose.bones['Bone'].location[2] = blind_up
store_armatures[6].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)

# ... Frame 300 (9 29/30 s)
C.scene.frame_set(frame = 300)
C.scene.camera.keyframe_insert(data_path = "location")
C.scene.camera.keyframe_insert(data_path = "rotation_euler")
store_armatures[7].pose.bones['Bone'].location[2] = blind_up
store_armatures[7].pose.bones['Bone'].keyframe_insert(data_path = "location", index = 2)
		
# ... Test Frame
C.scene.frame_set(frame = 0)
C.scene.camera.location = loc0
C.scene.camera.keyframe_insert(data_path = "location")
C.scene.camera.rotation_euler = rot0
C.scene.camera.keyframe_insert(data_path = "rotation_euler")

# Set all camera fcurves to linear interpolation
for fcurve in C.scene.camera.animation_data.action.fcurves:
	for point in fcurve.keyframe_points:
		point.interpolation = 'LINEAR'

# Instead of rendering, save as a different blend file (only in background mode)
# NOTE: Otherwise Mist map was unwritten memory
if bpy.app.background:
	project_filepath = 'classroom/ClassroomVideo_{}_{}spp.blend'.format(viewid, nsamples)
	bpy.ops.wm.save_mainfile(filepath = project_filepath, check_existing = False)
	
	output_dir = 'classroom/{}_{}spp'.format(viewid, nsamples)
	if not os.path.exists(output_dir):
		os.makedirs(output_dir)
